"""Main file for running the program from the command line."""

from __future__ import annotations

from pathlib import Path
from shutil import rmtree

from .configuration import UserConfigs
from .long_runs import (
    generate_week_barplots,
    generate_wordclouds,
)
from .models import ConversationSet
from .utils import latest_bookmarklet_json


def main() -> None:
    """Run the program."""
    print(
        "Welcome to ChatGPT Data Visualizer ✨📊!\n\n"
        "Follow the instructions in the command line.\n\n"
        "Press 'ENTER' to select the default options.\n\n"
        "If you encounter any issues 🐛, please report 🚨 them here:\n\n"
        "➡️ https://github.com/mohamed-chs/chatgpt-history-export-to-md/issues/new/choose"
        " 🔗\n\n",
    )

    user = UserConfigs()

    user.prompt()

    print("\n\nAnd we're off! 🚀🚀🚀\n")

    user.set_model_configs()

    print("Loading data 📂 ...\n")

    entire_collection = ConversationSet.from_zip(user.lookup["zip_filepath"])

    bkmrklet_json = latest_bookmarklet_json()
    if bkmrklet_json:
        print("Found bookmarklet download, loading 📂 ...\n")
        bkmrklet_collection = ConversationSet.from_json(bkmrklet_json)
        entire_collection.update(bkmrklet_collection)

    output_folder = Path(user.lookup["output_folder"])

    # overwrite the output folder if it already exists (might change this in the future)
    if output_folder.exists() and output_folder.is_dir():
        rmtree(output_folder)

    output_folder.mkdir(parents=True, exist_ok=True)

    markdown_folder = output_folder / "Markdown"

    entire_collection.save(markdown_folder, progress_bar=True)

    print(f"\nDone ✅ ! Check the output 📄 here : {markdown_folder.as_uri()} 🔗\n")

    graph_folder = output_folder / "Graphs"
    graph_folder.mkdir(parents=True, exist_ok=True)

    generate_week_barplots(
        entire_collection,
        graph_folder,
        **user.lookup["graph"],
        progress_bar=True,
    )

    print(f"\nDone ✅ ! Check the output 📈 here : {graph_folder.as_uri()} 🔗\n")
    print("(more graphs 📈 will be added in the future ...)\n")

    wordcloud_folder = output_folder / "Word Clouds"
    wordcloud_folder.mkdir(parents=True, exist_ok=True)

    generate_wordclouds(
        entire_collection,
        wordcloud_folder,
        **user.lookup["wordcloud"],
        progress_bar=True,
    )

    print(f"\nDone ✅ ! Check the output 🔡☁️ here : {wordcloud_folder.as_uri()} 🔗\n")

    print("Writing custom instructions 📝 ...\n")

    cstm_inst_filepath = output_folder / "custom_instructions.json"

    entire_collection.save_custom_instructions(cstm_inst_filepath)

    print(f"\nDone ✅ ! Check the output 📝 here : {cstm_inst_filepath.as_uri()} 🔗\n")

    print(
        "ALL DONE 🎉🎉🎉 !\n\n"
        f"Explore the full gallery 🖼️ at: {output_folder.as_uri()} 🔗\n\n"
        "I hope you enjoy the outcome 🤞.\n\n"
        "If you appreciate it, kindly give the project a star 🌟 on GitHub :\n\n"
        "➡️ https://github.com/mohamed-chs/chatgpt-history-export-to-md 🔗\n\n",
    )



"""Module for handling user configuration and updating the models."""

from __future__ import annotations

from questionary import (
    Choice,
    Style,
    checkbox,
    select,
)
from questionary import (
    path as qst_path,
)
from questionary import (
    text as qst_text,
)

from .models import Conversation, Message
from .utils import (
    DEFAULT_USER_CONFIGS,
    colormaps,
    font_names,
    font_path,
    stem,
    validate_header,
    validate_zip,
)

CUSTOM_STYLE = Style(
    [
        ("qmark", "fg:#34eb9b bold"),
        ("question", "bold fg:#e0e0e0"),
        ("answer", "fg:#34ebeb bold"),
        ("pointer", "fg:#e834eb bold"),
        ("highlighted", "fg:#349ceb bold"),
        ("selected", "fg:#34ebeb"),
        ("separator", "fg:#eb3434"),
        ("instruction", "fg:#eb9434"),
        ("text", "fg:#b2eb34"),
        ("disabled", "fg:#858585 italic"),
    ],
)


class UserConfigs:
    """Class for handling user configuration."""

    def __init__(self) -> None:
        """Initialize UserConfigs object."""
        # will implement a way to read from a config file later ...

        self.lookup = DEFAULT_USER_CONFIGS.copy()

    def prompt(self) -> None:
        """Prompt the user for input and update the configs."""
        lookup = self.lookup

        lookup["zip_filepath"] = qst_path(
            "Enter the path to the zip file :",
            lookup["zip_filepath"],
            validate=validate_zip,
            style=CUSTOM_STYLE,
        ).ask()

        lookup["output_folder"] = qst_path(
            "Enter the path to the output folder :",
            lookup["output_folder"],
            style=CUSTOM_STYLE,
        ).ask()

        for author_role in lookup["message"]["author_headers"]:
            lookup["message"]["author_headers"][author_role] = qst_text(
                f"Enter the message header (#) for messages from '{author_role}' :",
                lookup["message"]["author_headers"][author_role],
                validate=validate_header,
                style=CUSTOM_STYLE,
            ).ask()

        lookup["conversation"]["markdown"]["latex_delimiters"] = select(
            "Select the LaTeX math delimiters you want to use :",
            ["default", "dollars"],
            lookup["conversation"]["markdown"]["latex_delimiters"],
            style=CUSTOM_STYLE,
        ).ask()

        yaml_choices = [
            Choice(title=header, checked=value)
            for header, value in lookup["conversation"]["yaml"].items()
        ]

        selected_headers = checkbox(
            "Select the YAML metadata headers you want to include :",
            yaml_choices,
            style=CUSTOM_STYLE,
        ).ask()

        for header in lookup["conversation"]["yaml"]:
            lookup["conversation"]["yaml"][header] = header in selected_headers

        font_name: str = select(
            "Select the font you want to use for the word clouds :",
            font_names(),
            stem(lookup["wordcloud"].get("font_path") or ""),
            style=CUSTOM_STYLE,
        ).ask()

        lookup["wordcloud"]["font_path"] = str(font_path(font_name))

        lookup["wordcloud"]["colormap"] = select(
            "Select the color theme you want to use for the word clouds :",
            colormaps(),
            lookup["wordcloud"].get("colormap"),
            style=CUSTOM_STYLE,
        ).ask()

        lookup["wordcloud"]["custom_stopwords"] = qst_text(
            "Enter custom stopwords (separated by commas) :",
            lookup["wordcloud"].get("custom_stopwords", ""),
            style=CUSTOM_STYLE,
        ).ask()

    def set_model_configs(self) -> None:
        """Set the configuration for all models."""
        Message.update_configs(self.lookup["message"])
        Conversation.update_configs(self.lookup["conversation"])

    # will implement a way to save to a config file later ...



"""Utility functions for the project."""

from __future__ import annotations

from pathlib import Path
from re import compile as re_compile
from re import sub as re_sub
from typing import Any, Literal, TypedDict
from zipfile import ZipFile

DOWNLOADS = Path.home() / "Downloads"


def latest_zip() -> Path:
    """Path to the most recently created zip file in the Downloads folder."""
    zip_files = list(DOWNLOADS.glob("*.zip"))

    if not zip_files:
        err_msg = f"No zip files found in {DOWNLOADS}"
        raise FileNotFoundError(err_msg)

    return max(zip_files, key=lambda x: x.stat().st_ctime)


def latest_bookmarklet_json() -> Path | None:
    """Path to the most recent JSON file in Downloads with 'bookmarklet' in the name."""
    bkmrklet_files = [x for x in DOWNLOADS.glob("*.json") if "bookmarklet" in x.name]

    if not bkmrklet_files:
        return None

    return max(bkmrklet_files, key=lambda x: x.stat().st_ctime)


def sanitize(filename: str) -> str:
    """Sanitized title of the conversation, compatible with file names."""
    anti_pattern = re_compile(r'[<>:"/\\|?*\n\r\t\f\v]+')

    return anti_pattern.sub("_", filename.strip()) or "untitled"


def close_code_blocks(text: str) -> str:
    """Ensure that all code blocks are closed."""
    # A code block can be opened with triple backticks, possibly followed by a lang name
    # It can only be closed however with triple backticks, with nothing else on the line

    open_code_block = False

    lines = text.split("\n")

    for line in lines:
        if line.startswith("```") and not open_code_block:
            open_code_block = True
            continue

        if line == "```" and open_code_block:
            open_code_block = False

    if open_code_block:
        text += "\n```"

    return text


def replace_latex_delimiters(text: str) -> str:
    """Replace all the LaTeX bracket delimiters in the string with dollar sign ones."""
    text = re_sub(r"\\\[", "$$", text)
    text = re_sub(r"\\\]", "$$", text)
    text = re_sub(r"\\\(", "$", text)

    return re_sub(r"\\\)", "$", text)


def stem(path: Path | str) -> str:
    """Return the `stem` of the given path."""
    return Path(path).stem


def root_dir() -> Path:
    """Path to the root directory of the project.

    might change when refactoring, currently it's `convoviz/`
    """
    return Path(__file__).parent


def font_names() -> list[str]:
    """List of font names in the `assets/fonts` folder."""
    fonts_path = root_dir() / "assets" / "fonts"
    return [font.stem for font in fonts_path.iterdir()]


def font_path(font_name: str) -> Path:
    """Path to the given font in the `assets/fonts` folder.

    `font_name` should be the stem of the font file, without the extension
    """
    return root_dir() / "assets" / "fonts" / f"{font_name}.ttf"


def default_font_path() -> Path:
    """Path to the default font in the `assets/fonts` folder."""
    return font_path("RobotoSlab-Thin")


def colormaps() -> list[str]:
    """List of colormaps in the `assets/colormaps.txt` file."""
    colormaps_path = root_dir() / "assets" / "colormaps.txt"
    with colormaps_path.open(encoding="utf-8") as file:
        return file.read().splitlines()


def validate_header(text: str) -> bool:
    """Return True if the given text is a valid markdown header."""
    max_header_level = 6
    return (
        1 <= text.count("#") <= max_header_level
        and text.startswith("#")
        and text[len(text.split()[0])] == " "
    )


def validate_zip(filepath: str | Path) -> bool:
    """Return True if the given path is a zip file with a `conversations.json` file."""
    filepath = Path(filepath)
    if not filepath.is_file() or filepath.suffix != ".zip":
        return False
    with ZipFile(filepath) as zip_ref:
        if "conversations.json" in zip_ref.namelist():
            return True
    return False


def get_archive(filepath: Path | str) -> Path:
    """Extract the zip and return the path to the extracted folder."""
    filepath = Path(filepath)
    folder = filepath.with_suffix("")

    with ZipFile(filepath) as file:
        file.extractall(folder)

    return folder


def code_block(text: str, lang: str = "python") -> str:
    """Wrap the given string in a code block."""
    return f"```{lang}\n{text}\n```"


# default configs


class AuthorHeaders(TypedDict):
    """Type hint for the author headers configs."""

    system: str
    user: str
    assistant: str
    tool: str


class MessageConfigs(TypedDict):
    """Type hint for the message configs."""

    author_headers: AuthorHeaders


DEFAULT_MESSAGE_CONFIGS: MessageConfigs = {
    "author_headers": {
        "system": "### System",
        "user": "# Me",
        "assistant": "# ChatGPT",
        "tool": "### Tool output",
    },
}


class MarkdownConfigs(TypedDict):
    """Type hint for the markdown configs."""

    latex_delimiters: Literal["default", "dollars"]


class YAMLConfigs(TypedDict):
    """Type hint for the yaml configs."""

    title: bool
    tags: bool
    chat_link: bool
    create_time: bool
    update_time: bool
    model: bool
    used_plugins: bool
    message_count: bool
    content_types: bool
    custom_instructions: bool


class ConversationConfigs(TypedDict):
    """Type hint for the conversation configs."""

    markdown: MarkdownConfigs
    yaml: YAMLConfigs


DEFAULT_CONVERSATION_CONFIGS: ConversationConfigs = {
    "markdown": {"latex_delimiters": "default"},
    "yaml": {
        "title": True,
        "tags": False,
        "chat_link": True,
        "create_time": True,
        "update_time": True,
        "model": True,
        "used_plugins": True,
        "message_count": True,
        "content_types": True,
        "custom_instructions": True,
    },
}


class GraphKwargs(TypedDict, total=False):
    """Type hint for the graph configs."""


class WordCloudKwargs(TypedDict, total=False):
    """Type hint for the wordcloud configs."""

    font_path: str
    colormap: str
    custom_stopwords: str
    background_color: str | None
    mode: Literal["RGB", "RGBA"]
    include_numbers: bool
    width: int
    height: int


DEFAULT_WORDCLOUD_CONFIGS: WordCloudKwargs = {
    "font_path": str(default_font_path()),
    "colormap": "prism",
    "custom_stopwords": "use, file, ",
    "background_color": None,
    "mode": "RGBA",
    "include_numbers": False,
    "width": 1000,
    "height": 1000,
}


class AllConfigs(TypedDict):
    """Type hint for the user config JSON file."""

    zip_filepath: str
    output_folder: str
    message: MessageConfigs
    conversation: ConversationConfigs
    wordcloud: WordCloudKwargs
    graph: GraphKwargs
    node: dict[str, Any]
    conversation_set: dict[str, Any]


DEFAULT_USER_CONFIGS: AllConfigs = {
    "zip_filepath": str(latest_zip()),
    "output_folder": str(Path.home() / "Documents" / "ChatGPT Data"),
    "message": DEFAULT_MESSAGE_CONFIGS,
    "conversation": DEFAULT_CONVERSATION_CONFIGS,
    "wordcloud": DEFAULT_WORDCLOUD_CONFIGS,
    "graph": {},
    "node": {},
    "conversation_set": {},
}



"""Node class and methods for the node object in a conversation.

object path : conversations.json -> conversation -> mapping -> mapping node

will implement methods to handle conversation branches, like
counting the number of branches,
get the branch of a given node,
and some other version control stuff
"""

from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict

from ._message import Message

if TYPE_CHECKING:
    from typing import Any, ClassVar

    from typing_extensions import NotRequired, Self

    from ._message import MessageJSON


class NodeJSON(TypedDict):
    """Type of a `node` in the `mapping` field of a `conversation`."""

    id: str
    message: NotRequired[MessageJSON | None]
    parent: NotRequired[str | None]
    children: list[str]


class Node:
    """Wrapper class for a `node` in the `mapping` field of a `conversation`.

    see `NodeJSON` and `models.Conversation` for more details
    """

    __configs: ClassVar[dict[str, Any]] = {}

    def __init__(self, node: NodeJSON) -> None:
        """Initialize Node object."""
        self.__data = node
        self._parent: Self | None = None
        self._children: list[Self] = []

    @classmethod
    def update_configs(cls, configs: dict[str, Any]) -> None:
        """Set the configuration for all nodes."""
        cls.__configs.update(configs)

    @property
    def n_id(self) -> str:
        """Get the id of the node."""
        return self.__data["id"]

    @property
    def message(self) -> Message | None:
        """Get the message of the node."""
        if "message" not in self.__data or self.__data["message"] is None:
            return None
        return Message(self.__data["message"])

    @property
    def parent(self) -> Self | None:
        """Get the parent of the node."""
        return self._parent

    @parent.setter
    def parent(self, node: Self) -> None:
        """Set the parent of the node."""
        self._parent = node

    @property
    def children(self) -> list[Self]:
        """Get the children of the node."""
        return self._children

    def add_child(self, node: Self) -> None:
        """Add a child to the node."""
        self.children.append(node)
        node.parent = self

    @classmethod
    def mapping(cls, mapping: dict[str, NodeJSON]) -> dict[str, Self]:
        """Return a dictionary of connected Node objects, based on the mapping."""
        nodes: dict[str, Self] = {}

        # First pass: Create nodes
        for key, value in mapping.items():
            nodes[key] = cls(value)

        # Second pass: Connect nodes
        for key, value in mapping.items():
            for child_id in value["children"]:
                nodes[key].add_child(nodes[child_id])

        return nodes

    @property
    def header(self) -> str:
        """Get the header of the node message, containing a link to its parent."""
        if self.message is None:
            return ""

        parent_link = (
            f"[parent ⬆️](#{self.parent.n_id})\n"
            if self.parent and self.parent.message
            else ""
        )
        return f"###### {self.n_id}\n{parent_link}{self.message.author_header}\n"

    @property
    def footer(self) -> str:
        """Get the footer of the node message, containing links to its children."""
        if len(self.children) == 0:
            return ""
        if len(self.children) == 1:
            return f"\n[child ⬇️](#{self.children[0].n_id})\n"

        footer = "\n" + " | ".join(
            f"[child {i+1} ⬇️](#{child.n_id})" for i, child in enumerate(self.children)
        )
        return footer + "\n"



"""Conversation model. Represents a single ChatGPT chat.

object path : conversations.json -> conversation (one of the list items)
"""

from __future__ import annotations

from datetime import datetime, timedelta, timezone
from json import load as json_load
from os import utime as os_utime
from pathlib import Path
from time import ctime
from typing import TYPE_CHECKING, TypedDict

from convoviz.data_analysis import generate_wordcloud
from convoviz.utils import (
    DEFAULT_CONVERSATION_CONFIGS,
    close_code_blocks,
    replace_latex_delimiters,
    sanitize,
)

from ._node import Node

if TYPE_CHECKING:
    from typing import Any

    from PIL.Image import Image
    from typing_extensions import Self, Unpack

    from convoviz.utils import ConversationConfigs, WordCloudKwargs

    from ._message import AuthorRole
    from ._node import NodeJSON


class ConversationJSON(TypedDict):
    """Type of a `conversation` in _a_ `json` file."""

    title: str
    create_time: float
    update_time: float
    mapping: dict[str, NodeJSON]
    moderation_results: list[Any]  # I'm yet to see this field populated :(
    current_node: str
    plugin_ids: list[str]
    conversation_id: str
    conversation_template_id: str
    id: str


class Conversation:
    """Wrapper class for a `conversation` in _a_ `json` file.

    see `ConversationJSON` for more details
    """

    __configs = DEFAULT_CONVERSATION_CONFIGS

    def __init__(self, conversation: ConversationJSON) -> None:
        """Initialize Conversation object."""
        self.__data = conversation

    @classmethod
    def update_configs(cls, configs: ConversationConfigs) -> None:
        """Set the configuration for all conversations."""
        cls.__configs.update(configs)

    @classmethod
    def from_json(cls, filepath: Path | str) -> Self:
        """Load the conversation from a JSON file."""
        filepath = Path(filepath)

        with filepath.open(encoding="utf-8") as file:
            return cls(json_load(file))

    @property
    def title(self) -> str:
        """Get the title of the conversation."""
        return self.__data["title"]

    @property
    def create_time(self) -> float:
        """Get the creation time of the conversation."""
        return self.__data["create_time"]

    @property
    def create_datetime(self) -> datetime:
        """Get the creation datetime of the conversation."""
        return datetime.fromtimestamp(self.create_time, timezone.utc)

    @property
    def update_time(self) -> float:
        """Get the update time of the conversation."""
        return self.__data["update_time"]

    @property
    def update_datetime(self) -> datetime:
        """Get the update datetime of the conversation."""
        return datetime.fromtimestamp(self.update_time, timezone.utc)

    @property
    def mapping(self) -> dict[str, Node]:
        """Get the mapping of the conversation."""
        return Node.mapping(self.__data["mapping"])

    @property
    def moderation_results(self) -> list[Any]:
        """Get the moderation results of the conversation."""
        return self.__data["moderation_results"]

    @property
    def current_node(self) -> Node:
        """Get the current node of the conversation."""
        return self.mapping[self.__data["current_node"]]

    @property
    def plugin_ids(self) -> list[str]:
        """Get the plugin ids of the conversation."""
        return self.__data["plugin_ids"]

    @property
    def conversation_id(self) -> str:
        """Get the conversation id of the conversation."""
        return self.__data["conversation_id"]

    @property
    def conversation_template_id(self) -> str:
        """Get the conversation template id of the conversation."""
        return self.__data["conversation_template_id"]

    @property
    def _all_message_nodes(self) -> list[Node]:
        """List of all nodes that have a message, including all branches."""
        return [node for node in self.mapping.values() if node.message]

    def _author_nodes(
        self,
        *authors: AuthorRole,
    ) -> list[Node]:
        """List of all nodes with the given author role (all branches)."""
        if len(authors) == 0:
            authors = ("user",)
        return [
            node
            for node in self._all_message_nodes
            if node.message and node.message.author_role in authors
        ]

    @property
    def leaf_count(self) -> int:
        """Return the number of leaves in the conversation."""
        return sum(1 for node in self._all_message_nodes if not node.children)

    @property
    def chat_link(self) -> str:
        """Chat URL."""
        return f"https://chat.openai.com/c/{self.conversation_id}"

    @property
    def content_types(self) -> list[str]:
        """List of all content types in the conversation (all branches)."""
        return list(
            {
                node.message.content_type
                for node in self._all_message_nodes
                if node.message
            },
        )

    def message_count(
        self,
        *authors: AuthorRole,
    ) -> int:
        """Return the number of 'user' and 'assistant' messages (all branches)."""
        if len(authors) == 0:
            authors = ("user",)
        return len(self._author_nodes(*authors))

    @property
    def model_slug(self) -> str | None:
        """ChatGPT model used for the conversation."""
        assistant_nodes: list[Node] = self._author_nodes("assistant")
        if not assistant_nodes:
            return None

        message = assistant_nodes[0].message

        return message.model_slug if message else None

    @property
    def used_plugins(self) -> list[str]:
        """List of all ChatGPT plugins used in the conversation."""
        return list(
            {
                node.message.metadata["invoked_plugin"]["namespace"]
                for node in self._author_nodes("tool")
                if node.message and "invoked_plugin" in node.message.metadata
            },
        )

    @property
    def custom_instructions(self) -> dict[str, str] | None:
        """Return custom instructions used for the conversation."""
        system_nodes = self._author_nodes("system")
        if len(system_nodes) < 2:
            return None

        context_message = system_nodes[1].message
        if context_message and context_message.metadata.get("is_user_system_message"):
            return context_message.metadata.get("user_context_message_data")

        return None

        # TODO: check if this is the same for conversations from the bookmarklet

    @property
    def yaml(self) -> str:
        """YAML metadata header for the conversation."""
        yaml_config = self.__configs["yaml"]

        yaml_map = {
            "title": self.title,
            "chat_link": self.chat_link,
            "create_time": ctime(self.create_time),
            "update_time": ctime(self.update_time),
            "model": self.model_slug,
            "used_plugins": self.used_plugins,
            "message_count": self.message_count("user", "assistant"),
            "content_types": self.content_types,
            "custom_instructions": self.custom_instructions,
        }

        yaml = ""

        for key, value in yaml_map.items():
            if yaml_config.get(key, True):
                yaml += f"{key}: {value}\n"

        if not yaml:
            return ""

        return f"---\n{yaml}---\n"

    @property
    def markdown(self) -> str:
        """Return the full markdown text content of the conversation."""
        markdown_config = self.__configs["markdown"]
        latex_delimiters = markdown_config["latex_delimiters"]

        markdown = self.yaml

        for node in self._all_message_nodes:
            if node.message:
                content = close_code_blocks(node.message.content_text)
                # prevent empty messages from taking up white space
                content = f"\n{content}\n" if content else ""
                if latex_delimiters == "dollars":
                    content = replace_latex_delimiters(content)
                markdown += f"\n{node.header}{content}{node.footer}\n---\n"

        return markdown

    def save(self, filepath: Path | str) -> None:
        """Save the conversation to the file, with added modification time."""
        filepath = Path(filepath)
        base_file_name = sanitize(filepath.stem)

        counter = 0
        while filepath.exists():
            counter += 1
            filepath = filepath.with_name(
                f"{base_file_name} ({counter}){filepath.suffix}",
            )

        with filepath.open("w", encoding="utf-8") as file:
            file.write(self.markdown)

        os_utime(filepath, (self.update_time, self.update_time))

    def timestamps(
        self,
        *authors: AuthorRole,
    ) -> list[float]:
        """List of all message timestamps from the given author role (all branches).

        Useful for generating time graphs.
        """
        if len(authors) == 0:
            authors = ("user",)
        return [
            node.message.create_time
            for node in self._author_nodes(*authors)
            if node.message and node.message.create_time
        ]

    def plaintext(
        self,
        *authors: AuthorRole,
    ) -> str:
        """Entire plain text from the given author role (all branches).

        Useful for generating word clouds.
        """
        if len(authors) == 0:
            authors = ("user",)
        return "\n".join(
            node.message.content_text
            for node in self._author_nodes(*authors)
            if node.message
        )

    def wordcloud(
        self,
        *authors: AuthorRole,
        **kwargs: Unpack[WordCloudKwargs],
    ) -> Image:
        """Generate a wordcloud from the conversation."""
        if len(authors) == 0:
            authors = ("user",)
        text = self.plaintext(*authors)
        return generate_wordcloud(text, **kwargs)

    @property
    def week_start(self) -> datetime:
        """Return the monday of the week the conversation was created in."""
        start_of_week = self.create_datetime - timedelta(
            days=self.create_datetime.weekday(),
        )

        return start_of_week.replace(hour=0, minute=0, second=0, microsecond=0)

    @property
    def month_start(self) -> datetime:
        """Return the first of the month the conversation was created in."""
        return datetime(
            self.create_datetime.year,
            self.create_datetime.month,
            1,
            tzinfo=timezone.utc,
        )

    @property
    def year_start(self) -> datetime:
        """Return the first of January of the year the conversation was created in."""
        return datetime(self.create_datetime.year, 1, 1, tzinfo=timezone.utc)



"""Represents a single message in a conversation. It's contained in a Node object.

object path : conversations.json -> conversation -> mapping -> mapping node -> message
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal, TypedDict

from convoviz.utils import DEFAULT_MESSAGE_CONFIGS, code_block

if TYPE_CHECKING:
    from typing import Any

    from typing_extensions import NotRequired

    from convoviz.utils import MessageConfigs


AuthorRole = Literal["user", "assistant", "system", "tool"]


class AuthorJSON(TypedDict):
    """Type of the `author` field in a `message`."""

    role: AuthorRole
    name: NotRequired[str | None]
    metadata: dict[str, Any]


class ContentJSON(TypedDict):
    """Type of the `content` field in a `message`."""

    content_type: str
    parts: NotRequired[list[str]]
    text: NotRequired[str]
    result: NotRequired[str]


class MetadataJSON(TypedDict):
    """Type of the `metadata` field in a `message`."""

    model_slug: NotRequired[str]
    invoked_plugin: NotRequired[dict[str, Any]]
    user_context_message_data: NotRequired[dict[str, Any]]


class MessageJSON(TypedDict):
    """Type of the `message` field in a `node`."""

    id: str
    author: AuthorJSON
    create_time: NotRequired[float | None]
    update_time: NotRequired[float | None]
    content: ContentJSON
    status: str
    end_turn: NotRequired[bool]
    weight: float
    metadata: MetadataJSON
    recipient: str


class Message:
    """Wrapper class for the `message` field in a `node`.

    see `MessageJSON` and `models.Node` for more details
    """

    __configs = DEFAULT_MESSAGE_CONFIGS

    def __init__(self, message: MessageJSON) -> None:
        """Initialize Message object."""
        self.__data = message

    @classmethod
    def update_configs(cls, configs: MessageConfigs) -> None:
        """Set the configuration for all messages."""
        cls.__configs.update(configs)

    @property
    def m_id(self) -> str:
        """Get the id of the message."""
        return self.__data["id"]

    @property
    def author(self) -> AuthorJSON:
        """Get the author of the message."""
        return self.__data["author"]

    @property
    def create_time(self) -> float | None:
        """Get the creation time of the message."""
        return self.__data.get("create_time")

    @property
    def update_time(self) -> float | None:
        """Get the update time of the message."""
        return self.__data.get("update_time")

    @property
    def content(self) -> ContentJSON:
        """Get the content of the message."""
        return self.__data["content"]

    @property
    def status(self) -> str:
        """Get the status of the message."""
        return self.__data["status"]

    @property
    def end_turn(self) -> bool | None:
        """Get the end_turn of the message."""
        return self.__data.get("end_turn")

    @property
    def weight(self) -> float:
        """Get the weight of the message."""
        return self.__data["weight"]

    @property
    def metadata(self) -> MetadataJSON:
        """Get the metadata of the message."""
        return self.__data["metadata"]

    @property
    def recipient(self) -> str:
        """Get the recipient of the message."""
        return self.__data["recipient"]

    @property
    def author_role(self) -> AuthorRole:
        """Return the role of the author of the message."""
        return self.author["role"]

    @property
    def author_header(self) -> str:
        """Get the title header of the message based on the configs."""
        return self.__configs["author_headers"][self.author_role]

    @property
    def content_text(self) -> str:
        """Get the text content of the message."""
        if "parts" in self.content:
            return str(self.content["parts"][0])
        if "text" in self.content:
            return code_block(self.content["text"])
        if "result" in self.content:
            return self.content["result"]

        # this error caught a very hidden bug in the data
        # I need more of these
        err_msg = f"No valid content found in message {self.m_id}"
        raise ValueError(err_msg)

    @property
    def content_type(self) -> str:
        """Get the content type of the message."""
        return self.content["content_type"]

    @property
    def model_slug(self) -> str | None:
        """Get the model used for the message."""
        return self.metadata.get("model_slug")



"""Models for the application."""

from ._conversation import Conversation, ConversationJSON
from ._conversation_set import ConversationSet
from ._message import Message, MessageJSON
from ._node import Node, NodeJSON

__all__ = [
    "Conversation",
    "ConversationJSON",
    "ConversationSet",
    "Message",
    "MessageJSON",
    "Node",
    "NodeJSON",
]



"""Module for all the data visualizations.

Should ideally only return matplotlib objects, and not deal with the filesystem.
"""

# pyright: reportUnknownMemberType = false

from __future__ import annotations

from collections import defaultdict
from datetime import datetime, timezone
from typing import TYPE_CHECKING

from matplotlib.figure import Figure
from nltk import download as nltk_download  # type: ignore[import-untyped]
from nltk.corpus import stopwords as nltk_stopwords  # type: ignore[import-untyped]
from nltk.data import find as nltk_find  # type: ignore[import-untyped]
from wordcloud import WordCloud  # type: ignore[import-untyped]

from .utils import DEFAULT_WORDCLOUD_CONFIGS

if TYPE_CHECKING:
    from matplotlib.axes import Axes
    from PIL.Image import Image
    from typing_extensions import Unpack

    from .utils import GraphKwargs, WordCloudKwargs


def generate_week_barplot(
    timestamps: list[float],
    title: str,
    **kwargs: Unpack[GraphKwargs],
) -> Figure:
    """Create a bar graph from the given timestamps, collapsed on one week."""
    dates = [datetime.fromtimestamp(ts, timezone.utc) for ts in timestamps]

    weekday_counts: defaultdict[str, int] = defaultdict(int)
    days = [
        "Monday",
        "Tuesday",
        "Wednesday",
        "Thursday",
        "Friday",
        "Saturday",
        "Sunday",
    ]

    for date in dates:
        weekday_counts[days[date.weekday()]] += 1

    x = days
    y = [weekday_counts[day] for day in days]

    fig: Figure = Figure(dpi=300)
    ax: Axes = fig.add_subplot()

    ax.bar(x, y)
    ax.set_xlabel("Weekday")
    ax.set_ylabel("Prompt Count")

    ax.set_title(title)

    ax.set_xticks(x)
    ax.set_xticklabels(x, rotation=45)
    fig.tight_layout()

    return fig


# Ensure that the stopwords are downloaded
def _load_nltk_stopwords() -> set[str]:
    """Load nltk stopwords."""
    try:
        nltk_find("corpora/stopwords")
    except LookupError:
        nltk_download("stopwords")

    languages = [
        "arabic",
        "english",
        "french",
        "german",
        "spanish",
        "portuguese",
    ]  # add more languages here ...

    return {word for lang in languages for word in nltk_stopwords.words(fileids=lang)}


def generate_wordcloud(
    text: str,
    **kwargs: Unpack[WordCloudKwargs],
) -> Image:
    """Create a wordcloud from the given text."""
    configs = DEFAULT_WORDCLOUD_CONFIGS.copy()
    configs.update(kwargs)

    nltk_stopwords = _load_nltk_stopwords()

    custom_stopwords = configs.get("custom_stopwords")
    custom_stopwords_list = custom_stopwords.split(sep=",") if custom_stopwords else []
    custom_stopwords_list = [
        word.strip().lower() for word in custom_stopwords_list if word.strip()
    ]

    stopwords = nltk_stopwords.union(set(custom_stopwords_list))

    wordcloud = WordCloud(
        font_path=configs.get("font_path"),
        width=configs.get("width"),  # pyright: ignore[reportGeneralTypeIssues]
        height=configs.get("height"),  # pyright: ignore[reportGeneralTypeIssues]
        stopwords=stopwords,  # pyright: ignore[reportGeneralTypeIssues]
        background_color=configs.get("background_color"),  # pyright: ignore[reportGeneralTypeIssues]
        mode=configs.get("mode"),  # pyright: ignore[reportGeneralTypeIssues]
        colormap=configs.get("colormap"),
        include_numbers=configs.get("include_numbers"),  # pyright: ignore[reportGeneralTypeIssues]
    ).generate(text)

    img: Image = wordcloud.to_image()
    return img
